# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
""" Configuration file reader / writer
"""
import re
import os
from configparser import ConfigParser, ExtendedInterpolation


_IS_NUMBER = re.compile('^-?[0-9].*')
_IS_ENV_VAR = re.compile('\$\{(\w.*)?\}')


class EnvironmentNotFoundError(Exception):
    pass


class ExtendedEnvironmentInterpolation(ExtendedInterpolation):
    def __init__(self):
        items = os.environ.iteritems()
        self.environment = {k: v.replace('$', '$$') for k, v in items}
        self.klass = super(ExtendedEnvironmentInterpolation, self)

    def before_get(self, parser, section, option, value, defaults):
        defaults = self.environment
        defaults['HERE'] = '$${HERE}'
        if parser.filename:
            defaults['HERE'] = os.path.dirname(parser.filename)

        if defaults['HERE'] == '':
            defaults['HERE'] = os.curdir

        result = self.klass.before_get(parser, section, option, value,
                                       defaults)
        if '\n' in result:
            return [line for line in [self._unserialize(line)
                                      for line in result.split('\n')]
                    if line != '']
        return self._unserialize(result)

    def before_set(self, parser, section, option, value):
        result = self.klass.before_set(parser, section, option, value)
        return self._serialize(result)

    def _serialize(self, value):
        if isinstance(value, bool):
            value = str(value).lower()
        elif isinstance(value, (int, long)):
            value = str(value)
        elif isinstance(value, (list, tuple)):
            value = '\n'.join(['    %s' % line for line in value]).strip()
        else:
            value = str(value)
        return value

    def _unserialize(self, value):
        if not isinstance(value, basestring):
            # already converted
            return value

        value = value.strip()
        if _IS_NUMBER.match(value):
            try:
                return int(value)
            except ValueError:
                pass
        elif value.startswith('"') and value.endswith('"'):
            return value[1:-1]
        elif value.lower() in ('true', 'false'):
            return value.lower() == 'true'
        return value


class Config(ConfigParser):

    def __init__(self, filename):
        # let's read the file
        ConfigParser.__init__(self, **self._configparser_kwargs())
        if isinstance(filename, basestring):
            self.filename = filename
            self.read(filename)
        else:
            self.filename = None
            self.read_file(filename)

    def optionxform(self, option):
        return option

    def _read(self, fp, filename):
        # first pass
        ConfigParser._read(self, fp, filename)

        # let's expand it now if needed
        defaults = self.defaults()

        def _list(name):
            if name not in defaults:
                return []
            value = defaults[name]
            if not isinstance(value, list):
                value = [value]
            return value

        if 'extends' in defaults or 'overrides' in defaults:
            interpolate = self._interpolation.before_get

            for file_ in _list('extends'):
                file_ = interpolate(self, 'defaults', 'extends', file_, {})
                self._extend(file_)

            for file_ in _list('overrides'):
                file_ = interpolate(self, 'defaults', 'overrides', file_, {})
                self._extend(file_, override=True)

    def get_map(self, section=None):
        """returns a dict representing the config set"""
        if section:
            return dict(self.items(section))

        res = {}
        for section in self:
            for option, value in self[section].iteritems():
                option = '%s.%s' % (section, option)
                res[option] = value
        return res

    def mget(self, section, option):
        value = self.get(section, option)
        if not isinstance(value, list):
            value = [value]
        return value

    def _extend(self, filename, override=False):
        """Expand the config with another file."""
        if not os.path.isfile(filename):
            raise IOError('No such file: %s' % filename)
        parser = ConfigParser(**self._configparser_kwargs())
        parser.optionxform = lambda option: option
        parser.filename = filename
        parser.read([filename])
        serialize = self._interpolation._serialize

        for section in parser:
            if section in self:
                for option in parser[section]:
                    if option not in self[section] or override:
                        value = parser[section][option]
                        self[section][option] = serialize(value)
            else:
                self[section] = parser[section]

    def _configparser_kwargs(self):
        return {
            'interpolation': ExtendedEnvironmentInterpolation(),
            'comment_prefixes': ('#',),
        }

    def scan_args(self, parser, strip_prefixes=None):
        args = []

        # for each option in the parser we look for it in the config
        prefixes = ['DEFAULT']
        if strip_prefixes is not None:
            prefixes.extend(strip_prefixes)

        # building the list we have
        scanned = {}
        for key, value in self.get_map().items():
            # type conversion
            if isinstance(value, (list, tuple)):
                value = ','.join([str(v) for v in value])

            scanned[self._convert_key(key, prefixes)] = value

        # now trying to see if we have matches
        args = []

        for action in parser._actions:
            option = action.option_strings
            if '--help' in option or option == []:
                continue

            option = option[-1]
            if option in scanned:
                args.append(option)
                value = scanned[option]
                if not isinstance(value, bool):
                    args.append(str(value))

        return args

    def _convert_key(self, key, prefixes=None):
        if prefixes is None:
            prefixes = []

        for prefix in prefixes:
            if key.startswith('%s.' % prefix):
                key = key[len('%s.' % prefix):]
                break

        key = key.replace('.', '-')
        key = key.replace('_', '-')
        return '--' + key

    def as_args(self, strip_prefixes=None, omit_sections=None,
                omit_options=None):
        """Returns a list that can be passed to argparse or optparse.

        Each section/option is turned into "--section-option value"

        If the value is a boolean, the value will be omited.
        If the value is a sequence, it will be converted to a comma
        separated list

        Options:

        * strip_prefixes: a list of section names that will be stripped
          so the argument coverted as --option instead of --section-option

        * omit_sections: a list of section to ignore

        * omit_options: a list of options to ignore. Each option
          is provided as a 2-tuple (section, option)
        """
        args = []

        prefixes = ['DEFAULT']
        if strip_prefixes is not None:
            prefixes.extend(strip_prefixes)

        if omit_sections is None:
            omit_sections = []

        if omit_options is None:
            omit_options = []

        omit_options = ['%s.%s' % (sec, option)
                        for sec, option in omit_options]

        def _omit(key):
            if key in omit_options:
                return True

            for sec in omit_sections:
                if key.startswith('%s.' % sec):
                    return True
            return False

        for key, value in self.get_map().items():
            if _omit(key):
                continue

            args.append(self._convert_key(key, prefixes))

            # type conversion
            if isinstance(value, bool):
                continue
            elif isinstance(value, (list, tuple)):
                value = ','.join([str(v) for v in value])

            args.append(str(value))

        return args


class SettingsDict(dict):
    """A dict subclass with some extra helpers for dealing with app settings.

    This class extends the standard dictionary interface with some extra helper
    methods that are handy when dealing with application settings.  It expects
    the keys to be dotted setting names, where each component indicates one
    section in the settings heirarchy.  You get the following extras:

        * setdefaults:  copy any unset settings from another dict
        * getsection:   return a dict of settings for just one subsection

    """

    separator = "."

    def copy(self):
        """D.copy() -> a shallow copy of D.

        This overrides the default dict.copy method to ensure that the
        copy is also an instance of SettingsDict.
        """
        new_items = self.__class__()
        for k, v in self.iteritems():
            new_items[k] = v
        return new_items

    def getsection(self, section):
        """Get a dict for just one sub-section of the config.

        This method extracts all the keys belonging to the name section and
        returns those values in a dict.  The section name is removed from
        each key.  For example::

            >>> c = SettingsDict({"a.one": 1, "a.two": 2, "b.three": 3})
            >>> c.getsection("a")
            {"one": 1, "two", 2}
            >>>
            >>> c.getsection("b")
            {"three": 3}
            >>>
            >>> c.getsection("c")
            {}

        """
        section_items = self.__class__()
        # If the section is "" then get keys without a section.
        if not section:
            for key, value in self.iteritems():
                if self.separator not in key:
                    section_items[key] = value
        # Otherwise, get keys prefixed with that section name.
        else:
            prefix = section + self.separator
            for key, value in self.iteritems():
                if key.startswith(prefix):
                    section_items[key[len(prefix):]] = value
        return section_items

    def setdefaults(self, *args, **kwds):
        """Import unset keys from another dict.

        This method lets you update the dict using defaults from another
        dict and/or using keyword arguments.  It's like the standard update()
        method except that it doesn't overwrite existing keys.
        """
        for arg in args:
            if hasattr(arg, "keys"):
                for k in arg:
                    self.setdefault(k, arg[k])
            else:
                for k, v in arg:
                    self.setdefault(k, v)
        for k, v in kwds.iteritems():
            self.setdefault(k, v)



# ***** BEGIN LICENSE BLOCK *****
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file,
# You can obtain one at http://mozilla.org/MPL/2.0/.
# ***** END LICENSE BLOCK *****
import argparse
import unittest
import tempfile
import os
from StringIO import StringIO

from konfig import Config, SettingsDict


_FILE_ONE = """\
[DEFAULT]
extends = ${TEMPFILE}

[one]
foo = bar
num = -12
st = "o=k"
lines = 1
        two
        3

env = some ${__STUFF__}
location = ${HERE}

[two]
a = b
"""

_FILE_TWO = """\
[one]
foo = baz
bee = 1
two = "a"

[three]
more = stuff
location = ${HERE}
"""

_FILE_THREE = """\
[DEFAULT]
extends = no-no,no-no-no-no,no-no-no-no,theresnolimit

[one]
foo = bar
"""

_FILE_FOUR = """\
[global]
foo = bar
baz = bawlp

[auth]
a = b
c = d

[storage]
e = f
g = h

[multi:once]
storage.i = j
storage.k = l

[multi:thrice]
storage.i = jjj
storage.k = lll
"""

_EXTRA = """\
[some]
stuff = True

[other]
thing = ok
"""


_FILE_OVERRIDE = """\
[DEFAULT]
overrides = ${TEMPFILE}

[one]
foo = bar
"""

_FILE_ARGS = """\
[circus]
httpd = True
zmq_endpoint = http://ok

[other]
stuff = 10
thing = bleh

[bleh]
mew = 10

[mi]
log_level = DEBUG
log_output = stdout
daemon = True
pidfile = pid
"""


class ConfigTestCase(unittest.TestCase):

    def setUp(self):
        os.environ['__STUFF__'] = 'stuff'
        fp, filename = tempfile.mkstemp()
        f = os.fdopen(fp, 'w')
        f.write(_FILE_TWO)
        f.close()
        os.environ['TEMPFILE'] = filename
        self.file_one = StringIO(_FILE_ONE)
        self.file_two = filename
        self.file_three = StringIO(_FILE_THREE)
        self.file_override = StringIO(_FILE_OVERRIDE)
        self.file_args = StringIO(_FILE_ARGS)
        fp, filename = tempfile.mkstemp()
        f = os.fdopen(fp, 'w')
        f.write(_FILE_FOUR)
        f.close()
        self.file_four = filename

    def tearDown(self):
        if '__STUFF__' in os.environ:
            del os.environ['__STUFF__']
        os.remove(self.file_two)

    def test_reader(self):
        config = Config(self.file_one)

        # values conversion
        self.assertEquals(config.get('one', 'foo'), 'bar')
        self.assertEquals(config.get('one', 'num'), -12)
        self.assertEquals(config.get('one', 'st'), 'o=k')
        self.assertEquals(config.get('one', 'lines'), [1, 'two', 3])
        self.assertEquals(config.get('one', 'env'), 'some stuff')

        # getting a map
        map = config.get_map()
        self.assertEquals(map['one.foo'], 'bar')

        map = config.get_map('one')
        self.assertEquals(map['foo'], 'bar')

        del os.environ['__STUFF__']
        self.assertEquals(config.get('one', 'env'), 'some stuff')

        # extends
        self.assertEquals(config.get('three', 'more'), 'stuff')
        self.assertEquals(config.get('one', 'two'), 'a')

    def test_nofile(self):
        # if a user tries to use an inexistant file in extensions,
        # pops an error
        self.assertRaises(IOError, Config, self.file_three)

    def test_settings_dict_copy(self):
        settings = SettingsDict({"a.one": 1,
                                 "a.two": 2,
                                 "b.three": 3,
                                 "four": 4})
        new_settings = settings.copy()
        self.assertEqual(settings, new_settings)
        self.assertTrue(isinstance(new_settings, SettingsDict))

    def test_settings_dict_getsection(self):
        settings = SettingsDict({"a.one": 1,
                                 "a.two": 2,
                                 "b.three": 3,
                                 "four": 4})

        self.assertEquals(settings.getsection("a"), {"one": 1, "two": 2})
        self.assertEquals(settings.getsection("b"), {"three": 3})
        self.assertEquals(settings.getsection("c"), {})
        self.assertEquals(settings.getsection(""), {"four": 4})

    def test_settings_dict_setdefaults(self):
        settings = SettingsDict({"a.one": 1,
                                 "a.two": 2,
                                 "b.three": 3,
                                 "four": 4})

        settings.setdefaults({"a.two": "TWO", "a.five": 5, "new": "key"})
        self.assertEquals(settings.getsection("a"),
                          {"one": 1, "two": 2, "five": 5})
        self.assertEquals(settings.getsection("b"), {"three": 3})
        self.assertEquals(settings.getsection("c"), {})
        self.assertEquals(settings.getsection(""), {"four": 4, "new": "key"})

    def test_location_interpolation(self):
        config = Config(self.file_one)
        # file_one is a StringIO, so it has no location.
        self.assertEquals(config.get('one', 'location'), '${HERE}')
        # file_two is a real file, so it has a location.
        file_two_loc = os.path.dirname(self.file_two)
        self.assertEquals(config.get('three', 'location'), file_two_loc)

    def test_override_mode(self):
        config = Config(self.file_override)
        self.assertEquals(config.get('one', 'foo'), 'baz')
        self.assertEquals(config.get('three', 'more'), 'stuff')

    def test_as_args(self):
        config = Config(self.file_args)
        args = config.as_args(strip_prefixes=['circus'],
                              omit_sections=['bleh', 'mi'],
                              omit_options=[('other', 'thing')])

        wanted = ['--other-stuff', '10', '--httpd',
                  '--zmq-endpoint', 'http://ok']
        wanted.sort()
        args.sort()
        self.assertEqual(args, wanted)

        args = config.as_args(omit_sections=['bleh', 'mi'])
        wanted = ['--circus-zmq-endpoint', 'http://ok', '--other-thing',
                  'bleh', '--other-stuff', '10', '--circus-httpd']
        wanted.sort()
        args.sort()
        self.assertEqual(args, wanted)


        # it also works with an argparse parser
        parser = argparse.ArgumentParser(description='Run some watchers.')
        parser.add_argument('config', help='configuration file', nargs='?')

        parser.add_argument('-L', '--log-level', dest='loglevel')
        parser.add_argument('--log-output', dest='logoutput')
        parser.add_argument('--daemon', dest='daemonize', action='store_true')
        parser.add_argument('--pidfile', dest='pidfile')

        args = config.scan_args(parser, strip_prefixes=['mi'])
        wanted = ['--log-level', u'DEBUG', '--log-output', u'stdout',
                  '--daemon', '--pidfile', u'pid']
        args.sort()
        wanted.sort()
        self.assertEqual(args, wanted)



from setuptools import setup, find_packages


with open('README.rst') as f:
    README = f.read()


mplv2 = "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)"
classifiers = ["Programming Language :: Python", mplv2,
               "Development Status :: 5 - Production/Stable"]


setup(name='konfig',
      version='0.6',
      url='https://github.com/tarekziade/konfig',
      packages=find_packages(),
      long_description=README,
      description=("Yet Another Config Parser."),
      author="Tarek Ziade",
      author_email="tarek@mozilla.com",
      include_package_data=True,
      install_requires = [
        'configparser', 'argparse'
      ],
      zip_safe=False,
      classifiers=classifiers,
      )

